import json
import os
from pathlib import Path
import random
import shutil

import numpy as np

from texnomagic import common
from texnomagic.symbol import TexnoMagicSymbol


class TexnoMagicAlphabet:
    def __init__(self, path=None, name=None):
        self.path = path
        self.name = name
        self._symbols = None

    @property
    def info_path(self):
        return self.path / 'texno_alphabet.json'

    @property
    def symbols_path(self):
        return self.path / 'symbols'

    @property
    def handle(self):
        return common.name2handle(self.name)

    @property
    def symbols(self):
        if self._symbols is None:
            self.load_symbols()
        return self._symbols

    def load(self, path=None):
        if path:
            self.path = path

        assert self.path
        info = json.load(self.info_path.open())

        name = info.get('name')
        if not name:
            name = self.path.name
        self.name = name

        return self

    def load_symbols(self):
        self._symbols = []
        for symbol_info_path in self.symbols_path.glob('*/texno_symbol.json'):
            symbol = TexnoMagicSymbol()
            symbol.load(symbol_info_path.parent)
            self._symbols.append(symbol)
        self.sort_symbols()

    def sort_symbols(self):
        if not self._symbols:
            return
        known = []
        for core_symbol in common.CORE_SYMBOLS_ORDER:
            for symbol in self._symbols:
                if symbol.meaning == core_symbol:
                    known.append(symbol)
                    self._symbols.remove(symbol)
                    break
        self._symbols = known + self._symbols

    def save(self):
        os.makedirs(self.path, exist_ok=True)
        info = {
            'name': self.name,
        }
        return json.dump(info, self.info_path.open('w'))

    def save_new_symbol(self, symbol):
        assert symbol.name

        if self._symbols is None:
            self.load_symbols()

        symbol.path = self.symbols_path / common.name2fn(symbol.name)
        symbol.save()
        return self._symbols.insert(0, symbol)

    def export(self, out_path=None):
        """
        export alphabet into a zipfile
        """
        if not out_path:
            out_path = common.EXPORT_PATH

        ar_fn = self.path.name
        out_fn = out_path / ar_fn
        return shutil.make_archive(
            out_fn, 'zip',
            root_dir=self.path.parent,
            base_dir=self.path.name,
        )

    def train_models(self, all=False):
        """
        train symbol models with available drawings

        train only missing models by default, use all to (re-)train all
        """
        new, fail, old = [], [], []
        for symbol in self.symbols:
            if all or not symbol.model.ready:
                if symbol.model.train_symbol(symbol):
                    symbol.model.save()
                    new.append(symbol)
                else:
                    fail.append(symbol)
            else:
                old.append(symbol)
        return new, fail, old

    def calibrate(self):
        self.train_models(all=True)

    def scores(self, drawing):
        """
        recognize drawing using all symbol models

        return a list of (symbol, score) tuples ordered by score desc
        """
        s = [(symbol, symbol.model.score(drawing)) for symbol in self.symbols]
        s = sorted(s, key=lambda x: x[1], reverse=True)
        return s

    def recognize(self, drawing):
        s = self.scores(drawing)
        if not s:
            return None, -1
        _symbol, score = s[0]
        if score < common.MIN_SCORE:
            return None, score
        return s

    def check(self):
        """
        check alphabet for problems
        """
        warns = dict()

        def log_warn(key, val=1.0):
            if key in warns:
                n, v = warns[key]
                warns[key] = (n + 1, (v * n + val) / (n + 1))
            else:
                warns[key] = (1, val)

        for symbol in self.symbols:
            if not symbol.model.ready:
                log_warn(('warn', 'missing_model', symbol), -1)
            for drawing in symbol.drawings:
                scores = self.scores(drawing)
                rsymbol, rscore = scores[0]
                if rsymbol.meaning != symbol.meaning:
                    log_warn(('error', 'wrong_symbol', symbol, rsymbol), rscore)
                    prob = 'wrong_symbol'

                for rsy, rsc in scores[1:]:
                    # check scores for other symbols too
                    if rsc > 0.6:
                        lvl = 'warn'
                        if rsc > 0.8:
                            lvl = 'error'
                        log_warn((lvl,'high_score', symbol, rsy), rsc)

        results = {}
        for (level, prob, *args), (n, score) in warns.items():
            if prob == 'high_score':
                msg = "%s drawing got high score in %s: %s" % (args[0].meaning, args[1].meaning, score)
            elif prob == 'missing_model':
                msg = "%s symbol is missing model" % args[0].meaning
            else:
                msg = "%s drawing recognized as %s: %s" % (args[0].meaning, args[1].meaning, score)
            if n > 1:
                msg += "  (x%s)" % n
            if level not in results:
                results[level] = []
            results[level].append(msg)

        return results

    def get_symbol_by_name(self, name):
        for s in self.symbols:
            if s.name == name:
                return s
        return None

    def random_symbol(self, exclude=None):
        if exclude:
            symbols = [s for s in self.symbols if s != exclude]
        else:
            symbols = self.symbols
        if len(symbols) < 1:
            return None
        return random.choice(symbols)

    def stats(self, full=False):
        if full:
            return "%s: %s symbols @ %s" % (self.name, len(self.symbols), self.path)
        return "%s: %s symbols" % (self.name, len(self.symbols))

    def __repr__(self):
        return "<TexnoMagicAlphabet: %s>" % self.stats()



"""
TexnoMagic CLI
"""
import json
import sys

import click

from texnomagic import __version__
from texnomagic import lang
from texnomagic.abcs import TexnoMagicAlphabets
from texnomagic import mods
from texnomagic import server as server_


CONTEXT_SETTINGS = {
    'help_option_names': ['-h', '--help'],
}


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, message='%(version)s',
                      help="Show TexnoMagic version and exit.")
def cli():
    """
    TexnoMagic CLI
    """
    pass


@cli.command()
@click.argument('abc', nargs=-1)
@click.option('-n', '--names', is_flag=True,
              help="List names only (no details).")
@click.option('-f', '--full', is_flag=True,
              help="List everything including symbols.")
def list_abcs(abc, names, full):
    """
    List all/selected TexnoMagic alphabets.
    """
    abcs = TexnoMagicAlphabets()
    abcs.load()

    for tag, _abcs in abcs.abcs.items():
        if not names:
            print("%s %s:" % (len(_abcs), tag))
        for abc_ in _abcs:
            if abc and abc_.name not in abc:
                continue
            if names:
                print(abc_.name)
            else:
                print("%s" % abc_.stats(full=full))
            if full:
                for symbol in abc_.symbols:
                    if names:
                        print("  %s" % symbol.name)
                    else:
                        print("  %s" % symbol.stats(full=True))

    return 0


@cli.command()
@click.argument('abc', nargs=-1)
def check_abcs(abc):
    """
    Check all/selected alphabets for issues.
    """
    all_abcs = TexnoMagicAlphabets()
    all_abcs.load()

    for tag, _abcs in all_abcs.abcs.items():
        for abc_ in _abcs:
            if abc and abc_.name not in abc:
                continue
            print("CHECK %s" % abc_.stats(full=True))
            r = abc_.check()
            for level, msgs in sorted(r.items()):
                for msg in msgs:
                    print("%s: %s" % (level.upper(), msg))
    return 0


@cli.command()
@click.argument('abc', nargs=-1)
@click.option('-a', '--all', is_flag=True,
              help="Re-train all models. (default: only missing)")
def train_abcs(abc, all):
    """
    Train (missing) models for all/selected alphabets.
    """
    all_abcs = TexnoMagicAlphabets()
    all_abcs.load()

    first = True
    for tag, _abcs in all_abcs.abcs.items():
        for abc_ in _abcs:
            if abc and abc_.name not in abc:
                continue
            if first:
                first = False
            else:
                print()
            print("%s" % abc_.stats(full=True))
            new, fail, old = abc_.train_models(all=all)
            if new:
                print("TRAIN %s symbol models: %s" % (len(new), ", ".join([s.meaning for s in new])))
            if fail:
                print("FAIL %s symbol models: %s" % (len(fail), ", ".join([s.meaning for s in fail])))
            if old:
                print("ORIG %s symbol models: %s" % (len(old), ", ".join([s.meaning for s in old])))

    return 0


@cli.command()
@click.argument('spell', nargs=-1, required=True)
def spell(spell):
    """
    Parse TexnoMagic spell.
    """
    text = " ".join(spell)
    txlang = lang.TexnoMagicLanguage()
    try:
        out = txlang.parse(text)
    except Exception as e:
        print("parse error: %s" % e)
        return 42

    try:
        print(json.dumps(out, indent=4, sort_keys=False))
    except TypeError:
        print(out)
    return 0


@cli.command()
@click.argument('port', type=int, nargs=1, default=server_.DEFAULT_PORT)
def server(port):
    """
    Start TexnoMagic TCP server on PORT.
    """
    server_.serve(port=port)


@cli.command()
@click.argument('abc', nargs=1, required=True)
def flip_y(abc):
    """
    Flip Y axis for all symbols in alphabet.
    """
    abcs = TexnoMagicAlphabets()
    abcs.load()
    abc_ = abcs.get_abc_by_name(abc)
    if not abc_:
        print("alphabet not found: %s" % abc)
        return 20
    print(abc_)
    for symbol in abc_.symbols:
        print(symbol)
        for drawing in symbol.drawings:
            drawing.load_curves()
            drawing.flip_y_axis()
            drawing.save()


@cli.command()
def list_mods():
    """
    List online Words of Power mods from wop.mod.io.
    """
    for m in mods.get_online_mods():
        print("%s: %s  @ %s" % (m.name_id, m.name, m.profile_url))


@cli.command()
@click.argument('mod', nargs=-1)
@click.option('-a', '--all', is_flag=True,
              help="Download ALL mods. (default: only selected)")
def download_mods(mod, all):
    """
    Download Words of Power mods from wop.mod.io.
    """
    def do_dl(mod_):
        print("DOWNLOAD MOD: %s" % mod_)
        mod_.download()

    all_mods = mods.get_online_mods()
    if all:
        # all mods
        for mod in all_mods:
            do_dl(mod)
        return

    # selected mods
    for mod_id in mod:
        for m in all_mods:
            if all or m.name_id == mod_id or m.name == mod_id:
                do_dl(m)
                break
        else:
            print("mod not found - skipping: %s" % mod_id)


def main():
    cli()


if __name__ == '__main__':
    main()



import json
import numpy as np
import os
from pathlib import Path
import re


def name2fn(name):
    fn = re.sub(r'\s+', '-', name.lower())
    return fn


def name2handle(name):
    return ''.join(ch for ch in name.lower() if ch.isalnum())


def get_data_path():
    appdata = os.environ.get('APPDATA')
    if appdata:
        # windows system
        p = Path(appdata)
    else:
        # normal system :)
        p = Path.home() / '.local/share'
    return p / 'WordsOfPower'

MIN_SCORE = 0.6

DATA_PATH = get_data_path()
USER_DATA_PATH = DATA_PATH / 'user'
MODS_DATA_PATH = DATA_PATH / 'mods'
EXPORT_PATH = DATA_PATH / 'export'

ALPHABETS_DIR = 'alphabets'
ALPHABETS_PATHS = {
    'user': USER_DATA_PATH / ALPHABETS_DIR,
    'mods': MODS_DATA_PATH / ALPHABETS_DIR,
}

BUFFER_SIZE = 1024

CORE_SYMBOLS_ORDER = [
    # elements
    'fire',
    'ice',
    'lightning',
    'water',
    'air',
    'earth',
    'life',
    'death',
    # shapes
    'bolt',
    'ball',
    'beam',
    'area',
    'cone',
    # targets
    'self',
    'friend',
    'enemy',
    # selectors
    'close',
    'far',
    'weak',
    'strong',
    'random',
    'all',
    # modifiers
    'big',
    'small',
    'fast',
    'slow',
    'homing',
]


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def int2bytes(x):
    return x.to_bytes(4, byteorder="little")


def bytes2int(b):
    return int.from_bytes(b, byteorder="little")



import csv
import itertools
import numpy as np


class TexnoMagicDrawing:

    def __init__(self, path=None, curves=None, points_range=1000.0):
        self.path = path
        self.points_range = points_range
        self._curves = None
        self._points = None
        if curves:
            self.set_curves(curves)

    @property
    def curves(self):
        if self._curves is None:
            self.load_curves()
        return self._curves

    @property
    def points(self):
        if self._points is None:
            self.load_curves()
        return self._points

    @property
    def name(self):
        if self.path:
            return self.path.name
        return None

    def set_curves(self, curves):
        # keep all points in single continuous numpy array
        self._points = np.array(list(itertools.chain(*curves)), dtype=np.float64)
        self._curves = []
        i = 0
        for curve in curves:
            n = len(curve)
            # curves are numpy views into main points array
            cview = self._points[i:i+n]
            self._curves.append(cview)
            i += n

    def load(self, path=None):
        if path:
            self.path = path
        return self

    def load_curves(self):
        curves = []
        curve = []
        with self.path.open('r') as f:
            reader = csv.reader(f)
            for row in reader:
                if not row or '' in row:
                    curves.append(curve)
                    curve = []
                    continue
                point = list(map(float, row[:2]))
                curve.append(point)
        curves.append(curve)
        self.set_curves(curves)

    def save(self):
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with self.path.open('w', newline='') as f:
            writer = csv.writer(f)
            first = True
            for curve in self.curves:
                if first:
                    first = False
                else:
                    # curves separator
                    writer.writerow([None, None])
                writer.writerows(curve.tolist())

    def normalize(self):
        """
        normalize drawing points in-place into <0, self.points_range> range
        """
        # move to [0,0]
        self._points -= np.min(self._points, axis=0)
        # normalize
        k = self.points_range / np.max((np.max(np.max(self._points, axis=0)), 0.2))
        self._points *= k
        # center
        offset = (self.points_range - np.max(self._points, axis=0)) / 2
        self._points += offset

    def flip_y_axis(self):
        """
        flip drawing along Y axis in-place

        useful for compatibility with systems that use different Y axis sign
        """
        self._points[:,1] = self.points_range - self._points[:,1]

    def curves_fit_area(self, pos, size):
        """
        return curves scaled to fit area

        useful for drawing curves in UI
        """
        pos = np.array(pos)
        size = np.array(size)

        k = np.min(size) / self.points_range
        max_range = self.points_range * k

        offset = pos + (size - max_range) / 2

        scurves = []
        for curve in self.curves:
            if len(curve) > 0:
                scurve = curve * k + offset
            else:
                scurve = curve
            scurves.append(scurve)
        return scurves

    def delete(self):
        if not self.path or not self.path.exists():
            return
        self.path.unlink()

    def __repr__(self):
        return '<TexnoMagicDrawing @ %s: %d points in %d curves>' % (
            self.path, len(self._points), len(self._curves))



import json
import numpy as np
import os
import random
import time

from texnomagic import common
from texnomagic.drawing import TexnoMagicDrawing
from texnomagic.model import TexnoMagicSymbolModel


class TexnoMagicSymbol:
    def __init__(self, path=None, name=None, meaning=None):
        self.path = path
        self.name = name
        self.meaning = meaning
        self._drawings = None
        self._model = None

    @property
    def info_path(self):
        return self.path / 'texno_symbol.json'

    @property
    def drawings_path(self):
        return self.path / 'drawings'

    @property
    def model_path(self):
        return self.path / 'model'

    @property
    def handle(self):
        return common.name2handle(self.name)

    @property
    def model(self):
        if self._model is None:
            self.load_model()
        return self._model

    def load(self, path=None):
        if path:
            self.path = path

        assert self.path
        info = json.load(self.info_path.open())

        name = info.get('name')
        if not name:
            name = self.path.name
        self.name = name
        self.meaning = info.get('meaning')

        return self

    def load_drawings(self):
        self._drawings = []
        for drawing_path in self.drawings_path.glob('*'):
            drawing = TexnoMagicDrawing()
            drawing.load(drawing_path)
            self._drawings.append(drawing)

    def load_model(self):
        model = TexnoMagicSymbolModel(self.model_path)
        model.load()
        self._model = model

    def train_model(self, n_gauss=0):
        if not self._model:
            self._model = TexnoMagicSymbolModel(self.model_path)
        if n_gauss:
            self._model.n_gauss = n_gauss
        return self._model.train_symbol(self)

    def save(self):
        self.path.mkdir(parents=True, exist_ok=True)
        info = {
            'name': self.name,
            'meaning': self.meaning,
        }
        return json.dump(info, self.info_path.open('w'))

    def save_new_drawing(self, drawing):
        assert drawing

        if self._drawings is None:
            self.load_drawings()

        fn = "%s_%s.csv" % (common.name2fn(self.name), int(time.time() * 1000))
        drawing.path = self.drawings_path / fn
        drawing.save()
        return self._drawings.insert(0, drawing)

    @property
    def drawings(self):
        if self._drawings is None:
            self.load_drawings()
        return self._drawings

    def get_all_drawing_points(self):
        pp = [d.points for d in self.drawings]
        if pp:
            return np.concatenate(pp)
        return np.array([])

    def random_drawing(self):
        if self.drawings:
            return random.choice(self.drawings)
        return None

    def stats(self, full=False):
        if full:
            return '%s (%s): %s drawings @ %s' % (self.name, self.meaning, len(self.drawings), self.path)
        return '%s (%s)' % (self.name, self.meaning)

    def __repr__(self):
        return '<TexnoMagicSymbol %s>' % self.stats()




import glob
from pathlib import Path

from texnomagic.abc import TexnoMagicAlphabet
from texnomagic import common


class TexnoMagicAlphabets:
    def __init__(self, paths=None):
        self.paths = paths or common.ALPHABETS_PATHS
        self.abcs = {}

    def load(self):
        self.abcs = {}
        for tag, path in self.paths.items():
            self.abcs[tag] = get_alphabets(path)

    def get_abc_by_name(self, name):
        for _, abcs in self.abcs.items():
            for abc in abcs:
                if abc.name == name:
                    return abc
        return None

    def save_new_alphabet(self, abc, tag='user'):
        assert abc.name
        abc.path = self.paths[tag] / common.name2fn(abc.name)
        abc.save()
        self.abcs[tag].insert(0, abc)
        return abc

    def stats(self):
        stats = []
        for tag, abcs in self.abcs.items():
            stats.append('%d %s' % (len(abcs), tag))

        if not stats:
            stats = ['no alphabets found :(']

        return ", ".join(stats)

    def __repr__(self):
        return "<TexnoMagicAlphabets: %s>" % self.stats()


def get_alphabets(paths=None):
    paths = paths or common.ALPHABETS_PATHS
    abcs = []
    for abc_info_path in paths.glob('*/texno_alphabet.json'):
        abc = TexnoMagicAlphabet()
        abc.load(abc_info_path.parent)
        abcs.append(abc)
    return abcs



"""
TexnoMagic JSON-RPC over TCP server

This is a primary way to use TexnoMagic outside of Python.

You can start the server from terminal by invoking this module:

    python -m texnomagic.server

The server is using Godot Engine networking convention of 4 initial message
bytes marking the total message length and thus you can use Godot's native
networking or simply write a client of your own in a language of your choice.

Please see `client.py` for a reference implementation of a client.
"""
import json
import logging
import socketserver
import sys

from jsonrpcserver import dispatch

from texnomagic import __version__
from texnomagic import common
from texnomagic.abcs import TexnoMagicAlphabets
from texnomagic.lang import TexnoMagicLanguage
from texnomagic.drawing import TexnoMagicDrawing
# must be loaded in order for jsonrpc.dispatch() to work
from texnomagic import requests


LOG_FORMAT = '[TexnoMagic] %(message).64s'
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG, format=LOG_FORMAT)

DEFAULT_PORT = 6969


def serve(host='localhost', port=DEFAULT_PORT, abcs=None):
    """
    start TexnoMagic TCP server and serve forever
    """
    logging.info("START TexnoMagic TCP server %s on %s:%s ..." % (__version__, host, port))
    if not abcs:
        abcs = TexnoMagicAlphabets()
        abcs.load()

    with socketserver.TCPServer((host, port), TexnoMagicTCPHandler) as server:
        server.context = {
            'abcs': abcs,
            'lang': TexnoMagicLanguage()
        }
        logging.info("alphabets: %s" % abcs.stats())
        logging.info("server is RUNNING at %s:%s (CTRL+C to terminate)", host, port)
        try:
            server.serve_forever()
        except (KeyboardInterrupt, SystemExit):
            logging.info("server is SHUTTING DOWN, bye o/")


class TexnoMagicTCPHandler(socketserver.BaseRequestHandler):
    """
    TexnoMagic JSON-RPC over TCP request handler

    Individual requests are processed in requests.py
    """
    def handle(self):
        logging.info("NEW STREAM: %s", self.client_address)
        while True:
            # self.request is the TCP socket connected to the client
            try:
                size_raw = self.request.recv(4)
            except ConnectionResetError:
                logging.info("CLOSED connection by client")
                return
            except ConnectionAbortedError:
                logging.info("ABORTED connection")
                return

            l = len(size_raw)
            if l == 0:
                logging.info("DISCONNECT (0 bytes read)")
                return
            elif l != 4:
                logging.warning("TOO FEW BYTES: %s", l)
                return
            size = common.bytes2int(size_raw)
            i = 0
            data_raw = bytes()
            while i < size:
                rest = size - i
                n = min(rest, common.BUFFER_SIZE)
                data_raw += self.request.recv(n)
                i += n
            data = data_raw.decode('utf-8')
            # please see requests.py for individual requests' code
            response = dispatch(data, context=self.server.context)
            if response:
                self.send_data(response)

    def finish(self):
        logging.info("STREAM CLOSED: %s", self.client_address)

    def send_data(self, data):
        raw_data = bytes(data.encode('utf-8'))
        head = common.int2bytes(len(raw_data))
        payload = head + raw_data
        return self.request.sendall(payload)


if __name__ == "__main__":
    # primitive TexnoMagic server ggCLI
    args = sys.argv[1:]
    port = DEFAULT_PORT
    if args:
        port = int(args[0])
    serve(port=port)



"""
TexnoMagic JSON-RPC over TCP client reference implementation

This is only used for testing in TexnoMagic.

You can also look at wopeditor for a full-fledged TexnoMagic client
implementation in Godot Engine's GDScript:

https://github.com/texnoforge/wopeditor
"""
import json
import socket

from texnomagic import common


class TexnoMagicClient:
    """
    simple reference implementation of TexnoMagic TCP/JSONRPC client
    """
    def __init__(self, host='localhost', port=6969):
        self.host = host
        self.port = port
        self.sock = None
        self.buffer_size = 1024
        self.last_id = 0

    def connect(self):
        """
        connect to TexnoMagic TCP server
        """
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.host, self.port))

    def request(self, method, params=None):
        """
        send JSONRTC request over TCP and return response
        """
        self.last_id += 1
        # prepare JSONRTC payload
        if not params:
            params = []
        req = {
            'jsonrpc': '2.0',
            'method': method,
            'params': params,
            'id': self.last_id
        }
        # SEND request to server
        body = bytes(json.dumps(req), "utf-8")
        # first 4 bytes is the length of the message
        head = common.int2bytes(len(body))
        payload = head + body
        self.sock.sendall(payload)

        # RECEIVE response from server
        head = self.sock.recv(4)
        # first 4 bytes is the length of the message
        size = common.bytes2int(head)
        reply = str(self.sock.recv(size), "utf-8")
        return json.loads(reply)

    def close(self):
        """
        close the connection to TexnoMagic TCP server
        """
        self.sock.close()



"""
TexnoMagic / Words of Power mods interface

Use this to parse/download TexnoMagic alphabets as mods from https://wop.mod.io
"""
from io import BytesIO
import json
import requests
import zipfile

from texnomagic import common


MODIO_API_KEY = "2071808b470b7abc78fc289e5d3396d1"
MODIO_GAME_ID = 1986
MODIO_API_URL = "https://api.mod.io/v1"
MODIO_MODS_URL = "%s/games/%s/mods" % (MODIO_API_URL, MODIO_GAME_ID)


def get_online_mods():
    mods_data = get_online_mods_data()
    return data2mods(mods_data)


def get_online_mods_data():
    params = {
        'api_key': MODIO_API_KEY
    }
    r = requests.get(MODIO_MODS_URL, params=params)
    return json.loads(r.text)


def data2mods(data):
    mods = []
    for mod_data in data['data']:
        mod = WoPOnlineMod(mod_data)
        mods.append(mod)
    return mods


class WoPOnlineMod:
    data = {}
    name = None
    name_id = None
    filename = None
    profile_url = None
    binary_url = None
    logo_url = None

    def __init__(self, data=None):
        if data:
            self.load_from_data(data)

    def load_from_data(self, data):
        self.name = data.get('name')
        self.name_id = data.get('name_id')
        self.profile_url = data.get('profile_url')
        modfile = data.get('modfile', {})
        self.filename = modfile.get('filename')
        self.binary_url = modfile.get('download', {}).get('binary_url')
        self.logo_url = data.get('logo', {}).get('thumb_320x180')
        self.data = data

    def download(self, path=None):
        if not path:
            path = common.ALPHABETS_PATHS['mods']
        br = requests.get(self.binary_url)
        filebytes = BytesIO(br.content)
        zipf = zipfile.ZipFile(filebytes)
        zipf.extractall(path)

    def pretty_data(self):
        return json.dumps(self.data, indent=4, sort_keys=True)

    def __repr__(self):
        return '<WoPOnlineMod: %s @ %s>' % (self.name, self.profile_url)



"""
TexnoMagic Language grammar and parsing
"""
from parsimonious.grammar import Grammar
from parsimonious.nodes import NodeVisitor


# Parsing Expression Grammar (PEG) for TexnoMagic Language
TEXNO_MAGIC_GRAMMAR = """
spell = spell_bolt / spell_ball / spell_beam / spell_area / spell_cone / spell_self / spell_shield / spell_blink


spell_bolt = modable_effect bolt_mods "bolt" dir

spell_ball = modable_effect bolt_mods "ball" dir

spell_beam = modable_effect bolt_mods "beam" dir

spell_cone = modable_effect area_mods "cone" dir

spell_area = modable_effect area_mods "area"

spell_self = modable_effect "self"

spell_shield = modable_effect "shield" shield_shape

spell_blink = "air" ws "lightning" dir


modable_effect = effect_mods effect
effect = (element ws)+
element = "fire" / "ice" / "water" / "air" / "lightning" / "earth" / "life" / "death" / "magic"

area_mod = mod_size
area_mods = (area_mod ws)*
bolt_mod = mod_size / mod_speed / mod_homing
bolt_mods = (bolt_mod ws)*
effect_mod = mod_size / mod_speed
effect_mods = (effect_mod ws)*

mod_size = "big" / "small"
mod_speed = "fast" / "slow"
mod_homing = "homing"

dir = (ws direction)?
direction = "forward" / "enemy" / "friend" / "random"

shield_shape = (ws ("area" / "cone" / "self"))?


ws = " "+
"""


class TexnoMagicLanguage:
    def __init__(self):
        self.grammar = Grammar(TEXNO_MAGIC_GRAMMAR)

    def parse(self, text):
        """
        use this to parse TexnoMagic spell into its data representation
        """
        tree = self.grammar.parse(text)
        v = TexnoVisitor()
        output = v.visit(tree)
        return output


class TexnoVisitor(NodeVisitor):
    def visit_element(self, node, vc):
        return node.text

    def visit_effect(self, node, vc):
        elems = [e[0] for e in vc]
        return elems

    def visit_direction(self, node, vc):
        return node.text

    def visit_dir(self, node, vc):
        if vc:
            return vc[0][1]
        return 'default'

    def visit_mod_size(self, node, vc):
        if vc[0].text == 'small':
            v = 0.5
        else:  # big
            v = 2
        d = {'size': v}
        return d

    def visit_mod_speed(self, node, vc):
        if vc[0].text == 'slow':
            v = 0.5
        else:  # fast
            v = 2
        d = {'speed': v}
        return d

    def visit_mod_homing(self, node, vc):
        d = {'homing': 1}
        return d

    def visit_bolt_mod(self, node, vc):
        return vc[0]

    def visit_area_mod(self, node, vc):
        return vc[0]

    def visit_effect_mod(self, node, vc):
        return vc[0]

    def visit_bolt_mods(self, node, vc):
        mods = [c[0] for c in vc]
        mods = merge_mods(mods)
        d = {
            'spell_mods': mods,
        }
        return d

    def visit_area_mods(self, node, vc):
        mods = [c[0] for c in vc]
        mods = merge_mods(mods)
        d = {
            'spell_mods': mods,
        }
        return d

    def visit_modable_effect(self, node, vc):
        mods = [c[0] for c in vc[0]]
        mods = merge_mods(mods)
        mfx = {
            'effect': vc[1],
            'effect_mods': mods,
        }
        return mfx

    def visit_shield_shape(self, node, vc):
        shape = 'self'
        if vc:
            shape = vc[0][1][0].text
        return shape

    def visit_spell_bolt(self, node, vc):
        return boltlike_spell('bolt', node, vc)

    def visit_spell_ball(self, node, vc):
        return boltlike_spell('ball', node, vc)

    def visit_spell_beam(self, node, vc):
        return boltlike_spell('beam', node, vc)

    def visit_spell_cone(self, node, vc):
        return boltlike_spell('cone', node, vc)

    def visit_spell_area(self, node, vc):
        s = {'spell': 'area'}
        s.update(vc[0])
        s.update(vc[1])
        return s

    def visit_spell_self(self, node, vc):
        s = {'spell': 'self'}
        s.update(vc[0])
        return s

    def visit_spell_shield(self, node, vc):
        s = {
            'spell': 'shield',
            'shape': vc[2],
        }
        s.update(vc[0])
        return s

    def visit_spell_blink(self, node, vc):
        s = {
            'spell': 'blink',
            'direction': vc[3],
        }
        return s

    def visit_spell(self, node, vc):
        return vc[0]

    def generic_visit(self, node, vc):
        return vc or node


def boltlike_spell(name, node, vc):
    s = {'spell': name}
    s.update(vc[0])
    s.update(vc[1])
    s['direction'] = vc[3]
    return s


def merge_mods(mods):
    if not mods:
        return mods
    head, *rest = mods
    merged = head.copy()
    for m in rest:
        for key, val in m.items():
            if key in merged:
                merged[key] = merged[key] * val
            else:
                merged[key] = val
    return merged



from collections import Counter
import numpy as np
import json

from sklearn import mixture

# TODO: fix in PyInstaller upstream
# hidden import for PyInstaller
import sklearn.utils._weight_vector

from texnomagic.common import NumpyEncoder


class TexnoMagicSymbolModel:

    def __init__(self, path=None):
        self.path = path
        self.ready = False
        self.gmm = None
        self.n_gauss = 10
        self.score_avg = 0
        self.labels_avg = []

    @property
    def info_path(self):
        if self.path:
            return self.path / 'texno_model.json'
        return None

    def train_symbol(self, symbol):
        """
        train symbol model from its drawings
        """
        points = symbol.get_all_drawing_points()
        n_points = len(points)
        if n_points < 2 * self.n_gauss:
            # insufficient data
            return False

        self.train_GMM(points)

        score_sum = 0.0
        label_sums = np.zeros(self.gmm.n_components)
        for d in symbol.drawings:
            score_sum += self.gmm.score(d.points)
            labels = self.gmm.predict(d.points)
            labels_norm = normalize_labels(labels, self.gmm.n_components)
            label_sums += labels_norm

        self.labels_avg = label_sums / label_sums.sum()
        self.score_avg = score_sum / len(symbol.drawings)
        self.ready = True
        return True

    def train_GMM(self, data):
        """
        traing GMM model from data points
        """
        # thanks scikit-learn <3
        self.gmm = mixture.GaussianMixture(n_components=self.n_gauss)
        if data is None:
            return
        self.gmm.fit(data)

    def score(self, drawing):
        if not self.ready:
            return -1
        log_score = self.gmm.score(drawing.points)
        labels = self.gmm.predict(drawing.points)

        score = self.score_avg / log_score
        label_counts = normalize_labels(labels, self.n_gauss)
        label_diff = sum(abs(label_counts - self.labels_avg))
        label_k = max(1 - label_diff, 0.3)

        return score * label_k

    def get_preview(self):
        """
        return data for drawing a model preview
        """
        slen = 1000.0
        ssize = np.array([slen, slen])

        p = {
            'type': 'gmm',
            'components': []
        }
        comps = []
        if not self.gmm or not hasattr(self.gmm, 'means_'):
            return p

        means = self.gmm.means_
        covs = self.gmm.covariances_
        weights = self.gmm.weights_
        # draw covariances for each Gaussian
        for i, cov in enumerate(covs):
            # eigenvectors are magic
            v, w = np.linalg.eigh(cov)
            u = w[0] / np.linalg.norm(w[0])
            angle = np.arctan2(u[1], u[0])
            angle = 180 * angle / np.pi  # convert to degrees
            size = 2. * np.sqrt(2.) * np.sqrt(v)
            center = means[i, :2]
            weight = weights[i]
            comp = [center.tolist(), size.tolist(), angle, weight]
            comps.append(comp)
        p['components'] = comps
        return p

    def save(self):
        self.path.mkdir(parents=True, exist_ok=True)
        info = {
            'model_type': 'gmm',
            'n_gauss': self.n_gauss,
            'score_avg': self.score_avg,
            'labels_avg': self.labels_avg,
            'params': self.gmm._get_parameters()
        }
        return json.dump(info, self.info_path.open('w'), cls=NumpyEncoder, indent=2)

    def load(self, path=None):
        if path:
            self.path = path

        # reinit model before load
        self.__init__(path=self.path)

        if not self.info_path.exists():
            return False

        info = json.load(self.info_path.open())
        self.n_gauss = info['n_gauss']
        self.score_avg = info['score_avg']
        self.labels_avg = np.array(info['labels_avg'])
        self.gmm = mixture.GaussianMixture(n_components=self.n_gauss)
        params = [np.array(p) for p in info['params']]
        self.gmm._set_parameters(params)
        self.ready = True
        return True

    def __repr__(self):
        return '<TexnoMagicSymbolModel @ %s>' % self.path


def normalize_labels(labels, n):
    counts = dict(Counter(labels))
    for i in range(n):
        if i not in counts:
            counts[i] = 0
    counts = np.array([c for _, c in sorted(counts.items(), key=lambda x: x[0])])
    counts = counts / counts.sum()
    return counts



"""
TexnoMagic RPC handling logic

Individual functions in this module marked with @jsonrpcserver.method decorator
are used for Remote Procedure Calls (RPC) by the TexnoMagic server.
"""
from jsonrpcserver import method, Success

from texnomagic import __version__
from texnomagic.drawing import TexnoMagicDrawing
from texnomagic import mods


@method
def reload(context):
    context['abcs'].load()
    return Success(True)


@method
def spell(context, text):
    return Success(context['lang'].parse(text))


@method
def recognize(context, abc, curves):
    if not curves:
        return []
    _abc = context['abcs'].get_abc_by_name(name=abc)
    if not _abc:
        raise ValueError("requested alphabet isn't available: %s" % abc)

    drawing = TexnoMagicDrawing(curves=curves)
    drawing.normalize()
    symbol, score = _abc.recognize(drawing)
    r = {
        'symbol': symbol.name if symbol else None,
        'score': score,
    }
    return Success(r)


@method
def recognize_top(context, abc, curves, n=0):
    if not curves:
        return []
    _abc = context['abcs'].get_abc_by_name(name=abc)
    if not _abc:
        raise ValueError("requested alphabet isn't available: %s" % abc)

    drawing = TexnoMagicDrawing(curves=curves)
    drawing.normalize()
    symbols = _abc.scores(drawing)
    symbols = [s for s in symbols if s[1] > 0]
    if n:
        n = int(n)
        symbols = symbols[:n]
    return Success([(s.name, score) for (s, score) in symbols])


@method
def train_symbol(context, abc, symbol, n_gauss=0):
    _abc = context['abcs'].get_abc_by_name(name=abc)
    if not _abc:
        raise ValueError("requested alphabet isn't available: %s" % abc)
    _symbol = _abc.get_symbol_by_name(name=symbol)
    if not _symbol:
        raise ValueError("requested symbol isn't available: %s" % symbol)
    r = _symbol.train_model(n_gauss=n_gauss)
    assert(r)
    _symbol.model.save()
    return Success(True)


@method
def model_preview(context, abc, symbol):
    _abc = context['abcs'].get_abc_by_name(name=abc)
    if not _abc:
        raise ValueError("requested alphabet isn't available: %s" % abc)
    _symbol = _abc.get_symbol_by_name(name=symbol)
    if not _symbol:
        raise ValueError("requested symbol isn't available: %s" % symbol)
    model = _symbol.model
    if not model:
        return {}
    p = model.get_preview()
    return Success(p)


@method
def download_mod(context, mod):
    all_mods = mods.get_online_mods()
    for m in all_mods:
        if m.name_id == mod or m.name == mod:
            m.download()
            return Success(True)
    return Success(False)


@method
def export_abc(context, abc):
    _abc = context['abcs'].get_abc_by_name(name=abc)
    if not _abc:
        raise ValueError("requested alphabet isn't available: %s" % abc)
    return Success(_abc.export())


@method
def version(context):
    return Success(__version__)
